#!/usr/bin/python

#    Copyright 2012 Urban Airship
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Inspect an iOS or Android project for configuration errors"""

import sys
import argparse
import logging
import os
import support
import iossupport
import androidsupport
import package_info

from iossupport import VerificationInformation


def display_version():
    """Display the version number and usage if desired"""

    print("Version:%s" % package_info.current_version())


NO_CONFIG_PLIST_ERROR = """
No AirshipConfig.plist in app bundle. Without the configuration plist the
verification API call cannot be made. Parsing the app key and secret from
code is slated for a future release.
"""


class IosVerify(object):
    """ Verify the build with Urban Airship """

    def execute_xcode_verify(self):
        """Execute all build verification tasks"""

        # Setup logger
        log = logging.getLogger(support.UA_LOGGER)

        if self.version:
            display_version()
            return support.EXIT_SUCCESS

        iossupport.clean()

        if self.diagnostic:
            log.info("Setting up diagnostic logging")
            support.setup_diagnostic_logging()

        # Extract entitlement info from app
        log.info("Executing codesign system call")

        if self.application_path is None:
            log.error("Path to application is required. application_path must"
                      " be set in command line options")
            return support.EXIT_FAILURE

        success = iossupport.execute_codesign_system_call(
            self.application_path)
        if not success:
            log.error("Error in codesign system call")
            return support.EXIT_FAILURE

        # Parse entitlement plist output
        entitlements_plist_path = \
            support.get_value_from_config(support.ENTITLEMENTS_PLIST_PATH)

        entitlement_plist = iossupport.read_entitlement_plist_from_path(
            entitlements_plist_path)
        if entitlement_plist is None:
            log.error("Entitlement plist could not be read")
            return support.EXIT_FAILURE

        # Read the AirshipConfig from the app
        if self.plist is None:
            airship_config_plist = iossupport.extract_airship_config_from_app(
                self.application_path)
        else:
            airship_config_plist = \
                iossupport.extract_airship_config_from_app(
                    self.application_path, self.plist)

        # Store entitlement and config info
        self.verification_info = VerificationInformation(
            entitlement_plist, airship_config_plist)
        iossupport.log.debug("Verification:%s", self.verification_info)

        # Check local entitlement settings and app configuration
        if not self.verification_info.check_aps_environment():
            return support.EXIT_FAILURE

        # If there is no plist, there is no reason to verify anything
        # else. This is done after extracting the entitlements, but
        # before the matching API call. Extracting the information from code
        # is slated for a future release.
        if not airship_config_plist:
            if self.diagnostic:
                self.verification_info.log_diagnostic_information()
            log.error(NO_CONFIG_PLIST_ERROR)
            return support.EXIT_FAILURE

        # Make API verification request
        request = iossupport.get_verification_request(self.verification_info)
        response = iossupport.make_request_against_api(request)
        if response.error is not None:
            log.error("API verification request failed")
            return support.EXIT_FAILURE
        iossupport.log.debug("Response %s", response)

        # API response dependent tests
        apns_server = self.verification_info.check_apns_server(response.json)
        if not apns_server:
            log.error("APNS server is not configured properly")
            return support.EXIT_FAILURE

        bundle_id = self.verification_info.check_bundle_id(response.json)
        if not bundle_id:
            log.error("Bundle id mismatch")
            return support.EXIT_FAILURE

        if self.diagnostic:
            # Print diagnostic info, this has the side effect of printing
            # to a file
            self.verification_info.log_diagnostic_information()

        return support.EXIT_SUCCESS


class AndroidVerify(object):
    """Verify an Android Project"""

    def execute_android_verify(self):
        """Verify the android build"""

        # Setup logger
        log = logging.getLogger(support.UA_LOGGER)

        if self.version:
            display_version()
            return support.EXIT_SUCCESS

        if self.diagnostic:
            log.info("Setting up diagnostic logging")
            support.setup_diagnostic_logging()

        if not os.path.isdir(self.application_path):
            log.error("Path %s is not a directory." % self.application_path)
            return support.EXIT_FAILURE

        if self.manifest_path:
            manifest_element_tree = \
                androidsupport.parse_android_manifest(self.manifest_path)
        else:
            manifest_element_tree = \
                androidsupport.parse_manifest_in_directory(
                    self.application_path)

        # TODO change this so that it doesn't stop until all tests are run
        if manifest_element_tree is None:
            log.error("There is a problem with the AndroidManifest")
            return support.EXIT_FAILURE

        package_name = \
            androidsupport.package_name_from_manifest(manifest_element_tree)
        log.info("Package Name %s" % package_name)

        test_has_failed = list()

        isMissing = androidsupport.is_missing_uses_permissions(
            manifest_element_tree)
        if isMissing:
            test_has_failed.append(True)

        isMissing = androidsupport.is_missing_package_dependent_permissions(
            manifest_element_tree, package_name)
        if isMissing:
            test_has_failed.append(True)

        isMissing = androidsupport.is_missing_receiver_attributes(
            manifest_element_tree)
        if isMissing:
            test_has_failed.append(True)

        isMissing = androidsupport.is_missing_action_attributes(
            manifest_element_tree)
        if isMissing:
            test_has_failed.append(True)

        isMissing = androidsupport.is_missing_android_provider_attribute(
            manifest_element_tree, package_name)
        if isMissing:
            test_has_failed.append(True)

        # TODO possibly check proguard for annotations
        if not androidsupport.is_analytics_implemented(self.application_path):
            test_has_failed.append(True)

        if not androidsupport.is_takeoff_called(self.application_path):
            test_has_failed.append(True)

        airship_config_properties = androidsupport.read_properties_file(
            self.application_path)

        if airship_config_properties is None:
            test_has_failed.append(True)
            # If there are no config properties, a request cannot be made
            response = None
        else:
            request = androidsupport.get_verification_request(
                airship_config_properties)
            response = support.make_request_against_api(request)

        if response is None:
            test_has_failed.append(True)
        elif response.error is not None:
            test_has_failed.append(True)
        elif not androidsupport.is_airship_configured_properly(
            response.json,
            airship_config_properties,
            package_name):

            test_has_failed.append(True)

    # If something is missing in the test results, there was a problem
        if True in test_has_failed:
            log.error("Problems found in the Android Project")
            return support.EXIT_FAILURE
        else:
            log.info("Android project verified successfully")
            return support.EXIT_SUCCESS

APPLICATION_PATH_MESSAGE = """
Path to Project.app directory for iOS or root project directory for Android
"""

IOS_EPILOG = """
Extract the entitlements from a .app directory, parse them, and
check for errors. An API call is made to Urban Airship using the
configured key/secret pair to return API app settings for
comparison
"""


def ios_main():
    """Parses arguments and runs the uav-ios tool against an app"""

    log = logging.getLogger(support.UA_LOGGER)

    parser =\
        argparse.ArgumentParser(description="Verify a build with Urban Airship",
                                epilog=IOS_EPILOG)

    parser.add_argument('application_path', type=str, nargs="?",
                        help=APPLICATION_PATH_MESSAGE)

    parser.add_argument('-d', '--diagnostic', action="store_true",
                        help="Write out diagnostic information to a file")

    parser.add_argument('--plist', type=str,
                        help="Name of the configuration plist, "
                             "defaults to \"AirshipConfig.plist\" ")

    parser.add_argument('-v', '--version', action="store_true",
                        help="Display version number and exit")

    # Short circuit if there are no command line args
    if len(sys.argv) <= 1:
        display_version()
        parser.print_usage()
        return support.EXIT_SUCCESS

    # This appends the args to the object as instance variables
    verify = IosVerify()
    parser.parse_args(namespace=verify)

    success = verify.execute_xcode_verify()
    # There is the possibility of a file logger
    logging.shutdown()

    if success == support.EXIT_SUCCESS:
        log.debug("Successful verification")
        return support.EXIT_SUCCESS
    else:
        log.error("Failed verification")
        return support.EXIT_FAILURE

ANDROID_EPILOG = """
Parse the AndroidManifest.xml and look for required settings. Search
code files for proper implementation of analytics and references to takeOff. An
API call is made to Urban Airship using the key/secret pair to return API
app settings for comparison.
"""

ANDROID_MANIFEST_HELP = """
If your application has more than one AndroidManifest.xml file, use this flag
to pass the path of the manifest file to use for verification.
"""


def android_arg_parser():
    """Setup the arg parser for the Android tool"""

    parser = \
        argparse.ArgumentParser(description="Verify a build with Urban Airship",
                                epilog=ANDROID_EPILOG)

    parser.add_argument('application_path', type=str, nargs="?",
                        help=APPLICATION_PATH_MESSAGE)

    parser.add_argument('-d', '--diagnostic', action="store_true",
                        help="Write out diagnostic information to a file")

    parser.add_argument('-m', '--manifest_path', action="store",
                        help=ANDROID_MANIFEST_HELP)

    parser.add_argument('-v', '--version', action="store_true",
                        help="Display version number and exit")
    return parser


def android_main():
    """Parses arguments and runs the uav-android tool against an app"""

    log = logging.getLogger(support.UA_LOGGER)

    parser = android_arg_parser()

    # Short circuit if there are no command line args
    if len(sys.argv) <= 1:
        display_version()
        parser.print_usage()
        return support.EXIT_SUCCESS

    verify = AndroidVerify()
    parser.parse_args(namespace=verify)

    success = verify.execute_android_verify()
    # There is the possibility of a file logger
    logging.shutdown()
    if success == support.EXIT_SUCCESS:
        log.debug("Successful verification")
        return support.EXIT_SUCCESS
    else:
        log.error("Verification failed")
        return support.EXIT_FAILURE





#    Copyright 2012 Urban Airship
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


"""Supporting methods for verifying an Android build from the project root
"""

import os
import fnmatch
import re

import support

from xml.etree import ElementTree as ET
from collections import namedtuple
from string import Template


#logging
log_format = '%(asctime)s:%(levelname)s %(message)s'
# have to come up with a common name, instead of using the same logger
log = support.log

# Named tuple for search results
SearchResults = namedtuple("SearchResults", ["found", "not_found"])

#    For all of the ANDROID type values listed here, a lot of them need to be
#    namespaced. If you see android:name in the manifest, you'll need to call
#    manifest_attribute_key_with_namespace() to get the proper key value. This
#    simply prepends the key with the namespace, so it becomes
#    {http://schemas.android.com/apk/res/android}:key_name
#    which is what the XML library expects for the tree.get() command.
#    Example:
#    tree.get({namespace}:name) returns value associated with android:name

# Well known manifest namespace outlined here:
# http://developer.android.com/guide/topics/manifest/manifest-element.html
ANDROID_XMLNS = "http://schemas.android.com/apk/res/android"

# Take off method call
ANDROID_TAKE_OFF_METHOD_CALL = "takeOff"

# File names
ANDROID_MANIFEST_FILENAME = "AndroidManifest.xml"
ANDROID_AIRSHIP_PROPERTIES_FILENAME = "airshipconfig.properties"

# Airship Config file keys
ANDROID_CONFIG_DEV_APP_KEY = "developmentAppKey"
ANDROID_CONFIG_DEV_APP_SECRET_KEY = "developmentAppSecret"
ANDROID_CONFIG_PROD_APP_KEY = "productionAppKey"
ANDROID_CONFIG_PROD_APP_SECRET_KEY = "productionAppSecret"
ANDROID_CONFIG_TRANSPORT_KEY = "transport"
ANDROID_CONFIG_GCM_SENDER_KEY = "gcmSender"
ANDROID_CONFIG_IS_PROD_KEY = "inProduction"
ANDROID_CONFIG_ANALYTICS_ENABLED_KEY = "analyticsEnabled"

# Airship API response keys
UA_API_RESPONSE_PRODUCTION_KEY = "production"
UA_API_RESPONSE_PACKAGE_KEY = "android_package"
UA_API_RESPONSE_GCM_CONFIGURED = "gcm_is_configured"


# True/False string comparisons
STRING_TRUE = ["yes", "YES", "y", "Y", "Yes", "True", "true", "T"]
STRING_FALSE = ["no", "NO", "n", "N", "No", "False", "false", "F"]

# XML attribute keys
ANDROID_MANIFEST_PACKAGE_KEY = "package"
ANDROID_MANIFEST_PROTECTION_LEVEL_KEY = "protectionLevel"
ANDROID_MANIFEST_NAME_KEY = "name"
ANDROID_MANIFEST_PERMISSION_KEY = "permission"
ANDROID_MANIFEST_PROTECTION_LEVEL_SIGNATURE_KEY = "signature"
ANDROID_MANIFEST_PROVIDER_KEY = "provider"
ANDROID_MANIFEST_AUTHORITIES_KEY = "authorities"
ANDROID_MANIFEST_EXPORTED_KEY = "exported"
ANDROID_MANIFEST_MULTIPROCESS_KEY = "multiprocess"


# Android uses-permission declarations
ANDROID_USES_PERMISSION_INTERNET = "android.permission.INTERNET"
ANDROID_USES_PERMISSION_NETWORK_STATE = "android.permission.ACCESS_NETWORK_STATE"
ANDROID_USES_PERMISSION_VIBRATE = "android.permission.VIBRATE"
ANDROID_USES_PERMISSION_GET_ACCOUNTS = "android.permission.GET_ACCOUNTS"
ANDROID_USES_PERMISSION_WAKE_LOCK = "android.permission.WAKE_LOCK"
ANDROID_USES_PERMISSION_C2DM_RECEIVE = "com.google.android.c2dm.permission.RECEIVE"

# Android intent receivers
ANDROID_INTENT_RECEIVE = "com.google.android.c2dm.permission.RECEIVE"

# Optional not yet used in the script verification
ANDROID_PERMISSION_LOCATION = "android.permission.ACCESS_FINE_LOCATION"

# Required uses permissions as an iterable
_required_uses_permissions = [
    ANDROID_USES_PERMISSION_INTERNET,
    ANDROID_USES_PERMISSION_NETWORK_STATE,
    ANDROID_USES_PERMISSION_VIBRATE,
    ANDROID_USES_PERMISSION_GET_ACCOUNTS,
    ANDROID_USES_PERMISSION_WAKE_LOCK,
    ANDROID_USES_PERMISSION_C2DM_RECEIVE,
]

# Android permissions that related to messaging
ANDROID_PERMISSION_C2DM_SEND = "com.google.android.c2dm.permission.SEND"

# Intent permissions
ANDROID_INTENT_RECEIVE = "com.google.android.c2dm.intent.RECEIVE"
ANDROID_INTENT_REGISTRATION = "com.google.android.c2dm.intent.REGISTRATION"
ANDROID_INTENT_PACKAGE_REPLACED = "android.intent.action.PACKAGE_REPLACED"

# Iterable collection of intents, keep them in sync when adding new ones
_required_intents = [
    ANDROID_INTENT_RECEIVE,
    ANDROID_INTENT_REGISTRATION,
    ANDROID_INTENT_PACKAGE_REPLACED
]

# Package based permissions
#"com.urbanairship.push.sample.permission.C2D_MESSAGE
ANDROID_PACKAGE_BASED_PERMISSION_C2D_MESSAGE = "{package_name}.permission.C2D_MESSAGE"

# Receiver attributes
ANDROID_RECEIVER_CORE_RECEIVER = "com.urbanairship.CoreReceiver"
ANDROID_RECEIVER_GCM_PUSH_RECEIVER = "com.urbanairship.push.GCMPushReceiver"
ANDROID_RECEIVER_GCM_PUSH_RECEIVER_PERMISSION = "com.google.android.c2dm.permission.SEND"

# Iterable collection of receivers, keep them in sync
_required_receivers = [
    ANDROID_RECEIVER_CORE_RECEIVER,
    ANDROID_RECEIVER_GCM_PUSH_RECEIVER,
    ANDROID_RECEIVER_GCM_PUSH_RECEIVER_PERMISSION
]

# Content provider
ANDROID_CONTENT_PROVIDER_NAME = "com.urbanairship.UrbanAirshipProvider"
ANDROID_CONTENT_PROVIDER_AUTHORITIES = "{package_name}.urbanairship.provider"
ANDROID_CONTENT_PROVIDER_EXPORTED = "false"
ANDROID_CONTENT_PROVIDER_MULTIPROCESS = "multiprocess"

_required_provider_attributes = [
    ANDROID_CONTENT_PROVIDER_NAME,
    ANDROID_CONTENT_PROVIDER_AUTHORITIES,
    ANDROID_CONTENT_PROVIDER_EXPORTED,
    ANDROID_CONTENT_PROVIDER_MULTIPROCESS
]

MISSING_PROPERTY = """
Missing {property} property in properties file. Unable to create verification
request to use against the API. Check you airshipconfig.properties file for
errors"""


def get_verification_request(android_properties):
    """Get the verification request.

    :param android_properties: Dictionary of android properties
    :type key: dict
    :return: Request to make against the API
    :type key: urllib2.request
    """

    isProduction = android_properties.get(ANDROID_CONFIG_IS_PROD_KEY, None)
    if isProduction is None:
        log.info("airshipconfig.properties is missing isProduction key"
                 "Assuming development")
        isProduction = False

    try:
        if isProduction:
            appKey = android_properties[ANDROID_CONFIG_PROD_APP_KEY]
            appSecret = android_properties[ANDROID_CONFIG_PROD_APP_SECRET_KEY]
        else:
            appKey = android_properties[ANDROID_CONFIG_DEV_APP_KEY]
            appSecret = android_properties[ANDROID_CONFIG_DEV_APP_SECRET_KEY]
    except KeyError as error:
        log.error(MISSING_PROPERTY.format(property=error))
        return None

    return support.get_base_verification_request(appKey=appKey,
                                                 appSecret=appSecret)


def namespace_attribute_key(key, namespace=ANDROID_XMLNS):
    """Setup key with Android manifest namespace

    :param key: Key for required value
    :type key: str
    :param namespace: Namespace to use for key setup, defaults to ANDROID_XMLNS
    :type key: str
    :return: Key in the format {namespace}key
    :rtype: str
    """

    namespace_key_template = Template('{$namespace}$key')
    values = dict(namespace=namespace, key=key)
    return namespace_key_template.substitute(values)

def parse_manifest_in_directory(project_directory):
    """Recursively search the given directory for the AndroidManifest

    :param project_directory: Root directory of the project
    :return: Returns the first manifest found. If there are multiple manifests
        the result is one of those. Use parse_android_manifest for that
        situation.
    :rtype: xml.etree.ElementTree
    """

    path = find_file_from_dir(project_directory, ANDROID_MANIFEST_FILENAME)

    try:
        path = path.next()
        log.debug("AndroidManifest path {}".format(path))
    except StopIteration:
        log.error("No file name {} found in {}".format(ANDROID_MANIFEST_FILENAME,
                                                       project_directory))
        return None

    return parse_android_manifest(path)


def parse_android_manifest(manifest_path):
    """Parse the values of the android manifest

    :param manifest_path: Path to the root directory of the Android project
        to test

    :return: ElementTree with the contents of the AndroidManifest
    :rtype: xml.etree.ElementTree
    """

    try:
        tree = ET.parse(manifest_path)
    except ET.ParseError as parseError:
        log.error("Error parsing AndroidManifest {}".format(parseError.message))
        return None

    return tree


# Based on Stack Overflow answer, works well
def find_file_from_dir(base_dir, target_file):
    """Walk down from the root path, return all files with the filename

    :param base_dir: Base directory to start searching from.
    :type key: str
    :param target_file: file to search for
    :type key: str
    :return: Generator of files that match target_file
    :rtype: generator
    """

    for root, dirs, files in os.walk(base_dir, topdown=True):
        for filename in files:
            if fnmatch.fnmatch(filename, target_file):
                full_path_to_file = os.path.join(root, filename)
                yield full_path_to_file


def package_name_from_manifest(manifest_element_tree):
    """Extract the package name from the element tree representing the AndroidManifest

    :param manifest_element_tree: Tree for the AndroidManifest
    :type key: ElementTree
    :return: Package name
    :rtype: str
    """

    root = manifest_element_tree.getroot()
    return root.get(ANDROID_MANIFEST_PACKAGE_KEY)

MISSING_ATTRIBUTE = """
The AndroidManifest is missing the attribute
{missing_permission}
Please add the required permission to the manifest.
"""


def is_missing_uses_permissions(manifest_element_tree):
    """Check the permissions of type uses-permission

    Looking for similar xml:

    .. code-block:: xml

        <uses-permission android:name="android.permission.VIBRATE"/>
        <uses-permission android:name="com.google.android.c2dm.permission.RECEIVE"/>
        <uses-permission android:name="android.permission.ACCESS_NETWORK_STATE"/>
        <uses-permission android:name="android.permission.WAKE_LOCK"/>
        <uses-permission android:name="android.permission.INTERNET"/>
        <uses-permission android:name="android.permission.GET_ACCOUNTS"/>

    :param manifest_element_tree: Tree for the AndroidManifest
    :type key: ElementTree
    :returns: True if there are missing permissions, false otherwise
    :rtype: bool
    """

    root = manifest_element_tree.getroot()
    permissions_list = []
    for permission in root.iter("uses-permission"):
        permissions_list.append(permission.get(namespace_attribute_key(
            ANDROID_MANIFEST_NAME_KEY)))

    search_results = \
        search_iterable_for_values(permissions_list, _required_uses_permissions)

    msg = "{}<uses-permission android:name=\"{}\"/>"
    for result in search_results.found:
        log.info(msg.format("Found: ", result))
    for result in search_results.not_found:
        log.error(msg.format("Missing permission: ", result))

    if len(search_results.not_found) is not 0:
        return True
    else:
        return False


def is_missing_package_dependent_permissions(manifest_element_tree,
                                             package_name):
    """Check for permissions that require the package name

    :param manifest_element_tree: Tree for the AndroidManifest
    :type key: ElementTree
    :param package_name: Android package name
    :return: True if correct permissions are present, false if not
    :rtype: bool
    """

    root = manifest_element_tree.getroot()
    c2dm_permission = ANDROID_PACKAGE_BASED_PERMISSION_C2D_MESSAGE.format(
        package_name=package_name)
    missing_permissions = []

    package_based_permission = \
        namespace_attribute_key(ANDROID_MANIFEST_NAME_KEY)
    package_based_permission_is_present = False

    # Looking for the
    # <permission android:name "package.name" android:protectionLevel:"signature"
    for permission in root.iter("permission"):
        name = permission.get(package_based_permission)
        protection_level = permission.get(namespace_attribute_key(
            ANDROID_MANIFEST_PROTECTION_LEVEL_KEY))
        if name == c2dm_permission and \
           protection_level == ANDROID_MANIFEST_PROTECTION_LEVEL_SIGNATURE_KEY:
            package_based_permission_is_present = True
            break

    if package_based_permission_is_present is False:
        msg = "<permission android:name \"{}\" android:protectionLevel:\"signature\"/>".format(
            c2dm_permission)
        missing_permissions.append(msg)
        msg = MISSING_ATTRIBUTE.format(missing_permission=msg)
        log.error(msg)

    # Looking for
    # <uses-permission android:name="package.foo.permission.C2D_MESSAGE" />
    package_based_permission_is_present = False
    for permission in root.iter("uses-permission"):
        name = permission.get(package_based_permission)
        if name == c2dm_permission:
            package_based_permission_is_present = True
            break

    if package_based_permission_is_present is False:
        msg = "<uses-permission android:name \"{}\"/>".format(c2dm_permission)
        missing_permissions.append(msg)
        msg = MISSING_ATTRIBUTE.format(missing_permission=msg)
        log.error(msg)

    if len(missing_permissions) is not 0:
        return True
    else:
        return False


def is_missing_receiver_attributes(manifest_element_tree):
    """Check required application receiver attributes

    Looking for similar XML

    .. code-block:: xml

        <receiver android:name="com.urbanairship.push.GCMPushReceiver" android:permission="com.google.android.c2dm.permission.SEND">
        <receiver android:name="com.urbanairship.push.GCMPushReceiver" android:permission="com.google.android.c2dm.permission.SEND">
        <receiver android:name="com.google.android.c2dm.permission.SEND" />

    :param manifest_element_tree: Tree for the AndroidManifest
    :type key: ElementTree
    :return: True if required permissions are present, false if not
    :rtype: bool
    """

    # This is different than the majority of the settings.
    # There are two elements we are interested in, and one of the two elements
    # has two values that we need

    root = manifest_element_tree.getroot()
    found_receiver_attributes = []
    core_receiver = ANDROID_RECEIVER_CORE_RECEIVER
    # The xml has a namespace, each time you see android in the xml, it's really
    # {http://schemas.android.com/apk/res/android}
    name = namespace_attribute_key(ANDROID_MANIFEST_NAME_KEY)

    # Search for <receiver android:name="com.urbanairship.CoreReceiver" />,
    # add it to the found list
    for receiver_attrib in root.iter("receiver"):
        attribute = receiver_attrib.get(name)
        if attribute == core_receiver:
            found_receiver_attributes.append(attribute)

    # Search for <receiver android:name="com.urbanairship.push.GCMPushReceiver"
    # android:permission="com.google.android.c2dm.permission.SEND">
    # and check both values. Just add the android:name value if found to list
    # of found values
    gcm_push_receiver = ANDROID_RECEIVER_GCM_PUSH_RECEIVER
    target_permission = "<receiver android:name=\"com.urbanairship.push.GCMPushReceiver\" android:permission=\"com.google.android.c2dm.permission.SEND/\">"
    for receiver_attrib in root.iter("receiver"):
        value = receiver_attrib.get(name)
        if value == gcm_push_receiver:
            permission_send_key = namespace_attribute_key(
                ANDROID_MANIFEST_PERMISSION_KEY)
            permission_send = receiver_attrib.get(permission_send_key)
            if permission_send == ANDROID_PERMISSION_C2DM_SEND:
                log.info("Found: %s in manifest" %
                         target_permission)
                # Add both values here, so that you get a positive result if
                # and only if both values exist in the correct place
                found_receiver_attributes.append(permission_send)
                found_receiver_attributes.append(value)
                break

    search_results = search_iterable_for_values(found_receiver_attributes,
                                                _required_receivers)

    # Log results, formatted to look like XML
    core_receiver_xml = "{}<receiver android:name=\"{}\" />"
    receiver_xml = "{}<receiver android:name=\"{}\" android:permission=\"{}/\">"

    for result in search_results.found:
        if result == ANDROID_RECEIVER_GCM_PUSH_RECEIVER_PERMISSION:
            log.info(core_receiver_xml.format("Found: ", ANDROID_RECEIVER_GCM_PUSH_RECEIVER_PERMISSION))
        if result == ANDROID_RECEIVER_GCM_PUSH_RECEIVER:
            msg = receiver_xml.format("Found: ",
                                      ANDROID_RECEIVER_GCM_PUSH_RECEIVER,
                                      ANDROID_RECEIVER_GCM_PUSH_RECEIVER_PERMISSION)
            log.info(msg)

    for result in search_results.not_found:
        if result == ANDROID_RECEIVER_GCM_PUSH_RECEIVER_PERMISSION:
            log.error(core_receiver_xml.format("Missing: ", ANDROID_RECEIVER_GCM_PUSH_RECEIVER_PERMISSION))
        if result == ANDROID_RECEIVER_GCM_PUSH_RECEIVER:
            msg = receiver_xml.format("Missing: ",
                                      ANDROID_RECEIVER_GCM_PUSH_RECEIVER,
                                      ANDROID_RECEIVER_GCM_PUSH_RECEIVER_PERMISSION)
            log.error(msg)

    if len(search_results.not_found) is not 0:
        return True
    else:
        return False


def is_missing_action_attributes(manifest_element_tree):
    """Check required intent filters have the required actions

    Looking for the similar xml

    .. code-block:: xml

        <action android:name="com.google.android.c2dm.intent.REGISTRATION"/>
        <action android:name="android.intent.action.PACKAGE_REPLACED"/>
        <action android:name="com.google.android.c2dm.intent.RECEIVE"/>

    :param manifest_element_tree: Tree for the AndroidManifest
    :type key: ElementTree
    :return: True if filters are missing, false otherwise
    :rtype: bool
    """

    root = manifest_element_tree.getroot()
    action_values = []
    for attribute in root.iter("intent-filter"):
        for action in attribute.findall("action"):
            name = namespace_attribute_key(
                ANDROID_MANIFEST_NAME_KEY)
            value = action.get(name)
            if value is not None:
                action_values.append(value)

    search_results = \
        search_iterable_for_values(action_values, _required_intents)

    msg = "{}<action android:name=\"{}\"\>"
    for found in search_results.found:
        log.info(msg.format("Found: ", found))
    for not_found in search_results.not_found:
        log.error(msg.format("Missing intent action:", not_found))

    if len(search_results.not_found) is not 0:
        return True
    else:
        return False

MISSING_PROVIDER_FIX = """
<provider android:name="com.urbanairship.UrbanAirshipProvider"
    android:authorities="{package_name}.urbanairship.provider"
    android:exported="false"
    android:multiprocess="true" />
"""


def is_missing_android_provider_attribute(manifest_element_tree, package_name):
    """Check required provider attribute

    Checks for the 4 required attributes (name, authorities, exported, and
    multiprocess) and the corresponding values

    :param manifest_element_tree: Tree for the AndroidManifest
    :type key: ElementTree
    :param package_name: Package name for project
    :type key: str
    :return: True if provider is correct, false otherwise
    :rtype: bool
    """

    root = manifest_element_tree.getroot()
    provider_keys_values = dict()
    attribute_name = namespace_attribute_key(
        ANDROID_MANIFEST_NAME_KEY)

    for attribute in root.iter(ANDROID_MANIFEST_PROVIDER_KEY):
        name = attribute.get(attribute_name)
        if name == ANDROID_CONTENT_PROVIDER_NAME:

            value = attribute.get(namespace_attribute_key(
                ANDROID_MANIFEST_AUTHORITIES_KEY))
            provider_keys_values[ANDROID_MANIFEST_AUTHORITIES_KEY] = value

            value = attribute.get(namespace_attribute_key(
                ANDROID_MANIFEST_EXPORTED_KEY))
            provider_keys_values[ANDROID_MANIFEST_EXPORTED_KEY] = value

            value = attribute.get(
                namespace_attribute_key(ANDROID_MANIFEST_MULTIPROCESS_KEY))
            provider_keys_values[ANDROID_MANIFEST_MULTIPROCESS_KEY] = value

            break

    required_attributes = required_provider_attributes(package_name)
    if provider_keys_values == required_attributes:
        return False
    else:
        msg = "Missing permissions for provider, the permission should have " \
              "the following structure"
        log.error(msg)
        log.error(MISSING_PROVIDER_FIX.format(package_name=package_name))
        return True


def required_provider_attributes(package_name):
    """Creates the required attributes for the provider element

    The provider element has 4 attributes, name, authorities, exported and
    multiprocess. The value of the authorities attribute is dependent on the
    package name. This method returns the three values other than the name,
    since the name is required to get the three values.

    :param package_name: The package name used to create the authorities value
    :type key: str
    :return: Dictionary with the 3 attribute values
    :rtype: dict
    """

    required_values = dict()
    required_values[ANDROID_MANIFEST_EXPORTED_KEY] = "false"
    required_values[ANDROID_MANIFEST_MULTIPROCESS_KEY] = "true"
    authority = "{}.urbanairship.provider".format(package_name)
    required_values[ANDROID_MANIFEST_AUTHORITIES_KEY] = authority

    return required_values

PACKAGE_NAME_MISMATCH = """
"Package name in manifest {app} does not match {server} on UA servers"
"""


def is_airship_configured_properly(airship_api_response,
                                   airship_config_properties,
                                   application_package_name):
    """Compare the UA server configuration with airshipconfig.properties

    Also checks the package name in the AndroidManifest against the API
    response.

    :param airship_api_response: The json response from UA
    :type key: dict
    :param airship_config_properties: Properties from the
        airshipconfig.properties file
        
    :type key: dict
    :param application_package_name: Package name for the application
    :type key: str
    :return: True if app is configured properly, false otherwise
    :rtype: bool
    """

    tests = []
    server_package_name = airship_api_response[UA_API_RESPONSE_PACKAGE_KEY]
    if server_package_name != application_package_name:

        log.info(PACKAGE_NAME_MISMATCH.format(app=application_package_name,
                                              server=server_package_name))
        tests.append(False)
    else:
        log.info("Package name is configured on UA and matches")

    if airship_api_response[UA_API_RESPONSE_GCM_CONFIGURED] is False:
        log.error("GCM is not configured on Urban Airship servers")
        tests.append(False)
    else:
        log.info("GCM is configured on the UA servers")

    # XOR by converting False/True to 0/1 with bool(Bool)
    isServerProduction = airship_api_response[UA_API_RESPONSE_PRODUCTION_KEY]
    isAppProduction = airship_config_properties[ANDROID_CONFIG_IS_PROD_KEY]

    if bool(isServerProduction) ^ bool(isAppProduction):
        log.error("There is a configuration error for production/development")
        log.error("airshipconfig.properties inProduction is %s" %
                  isAppProduction)
        log.error("UA Server Production value is %s" % isServerProduction)
        tests.append(False)
    else:
        log.info("Server and application production values are correct")

    if False in tests:
        return False
    else:
        return True


def read_properties_file(root_directory):
    """Read the airshipconfig.properties file

    This assumes that there is a single airshipconfig.properties file, and it
    is somewhere in the project directory, or a subdirectory of the project
    directory. This stops at the first instance of a property file. The file
    is parsed with rudimentary string splitting based on the "=" character

    :param root_directory: Directory to search from
    :return: Contents of properties file, None on error
    :rtype: dict
    """

    path_iterator = find_file_from_dir(root_directory,
                                       ANDROID_AIRSHIP_PROPERTIES_FILENAME)
    try:
        airship_config_file = path_iterator.next()
    except StopIteration:
        airship_config_file = None

    if airship_config_file is None:
        log.error("No config file found. Searched for {} from root {}".format(
            ANDROID_AIRSHIP_PROPERTIES_FILENAME,  root_directory))
        return None
    else:
        properties = dict()
        for line in open(airship_config_file, "r"):
            key_value = line.strip().split("=")
            # Skip comments
            if "#" in key_value[0]:
                continue
            # Only go for key/value pairs
            if len(key_value) != 2:
                continue
            key = key_value[0].strip()
            value = key_value[1].strip()
            log.debug("Adding config properties key:{} val:{}".format(key, value))
            # strip keys and values of whitespace
            properties[key] = value
        # TODO this will throw an exception if the key is not present
        match = None
        try:
            match = re.match('[Yy]', properties[ANDROID_CONFIG_IS_PROD_KEY])
        except KeyError:
            log.info("Missing isProduction key in AirshipConfig.properties. "
                     "Defaulting to development.")
        if match is None:
            properties[ANDROID_CONFIG_IS_PROD_KEY] = False
        else:
            properties[ANDROID_CONFIG_IS_PROD_KEY] = True

    return properties


def search_iterable_for_values(values, targets):
    """Takes a collection of values an then searches it for all targets.

    :param values: Values to be searched
    :param targets: Values to look for in the collection

    :return: Tuple of values in the order of Found, Missing. Both
    :rtype: tuple
    """

    values = set(values)
    targets = set(targets)
    log.debug("Searching for\n%s\n in \n%s\n", targets, values)
    targets_in_values = values.intersection(targets)
    targets_missing_from_values = targets.difference(values)

    return SearchResults(found=targets_in_values,
                         not_found=targets_missing_from_values)


# TODO add support for manually implementing reports with onStart/onStop
def is_analytics_implemented(root_path):
    """Recursively search for java files that use Android reports

    Checks for classes that extend InstrumentedActivity or
    InstrumentedListActivity and if analytics are enabled. Will also return
    false if there is no config file, or there is a parse error

    :param root_path: Root path to begin search
    :return: True if classes exist an analytics are not disabled
    :rtype: bool
    """

    airship_config = read_properties_file(root_path)

    # Analytics is not enabled in this case, and everything is probably broken
    # anyway
    if airship_config is None:
        log.error("No airshipconfig.properties found at path %s" % root_path)
        return False

    is_analytics_enabled = True
    is_analytics_enabled_string = None

    # There is a good chance this key has been omitted from the config file
    try:
        is_analytics_enabled_string = \
            airship_config[ANDROID_CONFIG_ANALYTICS_ENABLED_KEY]
    except KeyError:
        log.info(
            "analyticsEnabled key not found in properties file, "
            "defaulting to true")

    if is_analytics_enabled_string is not None:
        if is_analytics_enabled_string in STRING_TRUE:
            is_analytics_enabled = True
        else:
            is_analytics_enabled = False

    if is_analytics_enabled is not True:
        log.info(
            "Analytics has been purposefully disabled in "
            "airshipconfig.properties")

        return False

    java_file_generator = find_file_from_dir(root_path, "*.java")

    files_with_analytics = list()
    instrumented_activity_pattern = re.compile("extends\s+InstrumentedActivity")
    instrumented_activity_list_pattern = \
        re.compile("extends\s+InstrumentedListActivity")

    for java_file_path in java_file_generator:

        with open(java_file_path, "r") as java_file:

            for line in java_file:

                if instrumented_activity_pattern.search(line) is not None:
                    log.info("Found instrumented activity at path %s" %
                             java_file_path)
                    files_with_analytics.append(java_file_path)

                if instrumented_activity_list_pattern.search(line) is not None:
                    log.info("Found instrumented list activity at path %s" %
                             java_file_path)
                    files_with_analytics.append(java_file_path)

    if len(files_with_analytics) == 0 and is_analytics_enabled is True:
        log.error("Analytics are enabled, but no files extend an "
                  "instrumented activity")

        return False

    return True


def is_takeoff_called(root_path):
    """Check to see if takeOff is called in code

    Checks for a class that extends Application and contains a call to
    takeOff: This will return true at the first match of a class that extends
    Activity and calls takeOff

    :param root_path: Recursively search from this directory
    :return: True if takeOff is present, false otherwise
    :rtype: bool
    """

    java_file_generator = find_file_from_dir(root_path, "*.java")
    take_off_found = False

    for java_file_path in java_file_generator:
        application_found = False

        # Look for java class that extends Application
        with open(java_file_path, "r") as java_file:
            pattern = re.compile("extends\s+Application")
            for line in java_file:
                if pattern.search(line) is not None:
                    log.info("Searching {} for call to takeOff".format(
                        java_file_path))
                    application_found = True
                    break

        # If the class extends Application, check for takeOff
        if application_found is True:
            with open(java_file_path, "r") as java_file:
                pattern = re.compile("takeOff")
                for line in java_file:
                    if pattern.search(line) is not None:
                        take_off_found = True
                        log.info("Found takeOff in file {} in line {}".format(
                            java_file_path, line.strip()))

        if take_off_found is True:
            break

    return take_off_found



#    Copyright 2012 Urban Airship
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

""" Support methods for verifying iOS builds
"""

import sys
import string
import subprocess
import plistlib
import os
import re
import support
import errno
import tempfile

from collections import namedtuple
from xml.parsers.expat import ExpatError

DEFAULT_CONFIG_PLIST_NAME = "AirshipConfig.plist"

# AirshipConfig.plist keys
APP_STORE_OR_AD_HOC_BUILD = "inProduction"
DEVELOPMENT_APP_KEY = "developmentAppKey"
DEVELOPMENT_APP_SECRET = "developmentAppSecret"
PRODUCTION_APP_SECRET = "productionAppSecret"
PRODUCTION_APP_KEY = "productionAppKey"

DEPRECATED_APP_STORE_OR_AD_HOC = "APP_STORE_OR_AD_HOC_BUILD"
DEPRECATED_DEVELOPMENT_APP_KEY = "DEVELOPMENT_APP_KEY"
DEPRECATED_DEVELOPMENT_APP_SECRET = "DEVELOPMENT_APP_SECRET"
DEPRECATED_PRODUCTION_APP_SECRET = "PRODUCTION_APP_SECRET"
DEPRECATED_PRODUCTION_APP_KEY = "PRODUCTION_APP_KEY"


# API JSON Response Keys, currently the API call returns the bundle id and
# a boolean indicating whether or not th app is in production
UA_API_RESPONSE_BUNDLE_ID_KEY = "bundle_id"
UA_API_RESPONSE_PRODUCTION_KEY = "production"

# Entitlement Plist keys
ENTITLE_APP_ID_KEY = "application-identifier"
ENTITLE_APS_ENV_KEY = "aps-environment"

# Filename for deprecated plist conversion output
CONVERTED_DEPRECATED_PLIST = "ConvertedAirshipConfig.plist"

# APNS/UA Sever names
PRODUCTION_SERVER = "production"
DEVELOPMENT_SERVER = "development"

#setup logging
log = support.log


def clean():
    """ Cleanup any previous build cruft

    Currently only erases previous entitlement plists
    """
    entitlement_plist_path = support.get_value_from_config(
        support.ENTITLEMENTS_PLIST_PATH)

    if os.path.exists(entitlement_plist_path):
        log.info("Removing old entitlements plist file at %s",
                 entitlement_plist_path)
        os.remove(entitlement_plist_path)


# API methods

APP_KEY_SECRET_INCORRECT_401 = """
The app key and secret pair are incorrect, or the app has not been setup on
Urban Airship. Please login to your account at go.urbanairship.com and review
the settings."""

APP_KEY_SECRET_INCORRECT_403 = """
The wrong value is being used for the app secret. This is the master secret,
and is used for API communication outside of the scope of a client app. It
should never be used in an app, or placed in an app bundle, as that would
be a security risk. Please review your app settings"""


def get_verification_request(ua_verification_information):
    """Get the verification request.

    The request is setup with
    basic auth headers parsed from the settings in AirshipConfig.plist
    parsed from the UAVerificationInformation passed in.

    :param ua_verification_information: VerificationInformation object
    with the data needed for this API request

    :type key: VerificationInformation
    :return: Pre authenticated http request
    :rtype: urllib2.Request
    """
    key, secret = ua_verification_information.key_secret()
    return support.get_base_verification_request(appKey=key, appSecret=secret)


def make_request_against_api(request):
    """Make a request against the UA API

    Returns a named tuple with response data as JSON and a possible error
    object which is either an HTTPError, 401,500, etc or a URLError which
    most likely occurs when there is no connection. All other errors will
    exception out. Will display the proper

    :param request: Request to make
    :type key: urllib2.Request
    :return: Named tuple in the form (response, error)
    :rtype: namedtuple
    """

    response = support.make_request_against_api(request)

    if response.error is not None:

        if response.error.code == 401:
            log.error(APP_KEY_SECRET_INCORRECT_401)

        if response.error.code == 403:
            log.error(APP_KEY_SECRET_INCORRECT_403)

    return response

DEPRECATED_PLIST_CONVERT_MSG = """
Plist parsing error, attempting to convert plist from old style plist
to XML plist
"""

AIRSHIP_CONFIG_MISSING_MSG = """
Error opening the plist at path {} with name {}, did you forget to
add it to the bundle?
"""


def extract_airship_config_from_app(application_path,
                                    plist_name=DEFAULT_CONFIG_PLIST_NAME):
    """ Attempts to read the AirshipConfig.plist out of th app bundle

    Makes the assumption the file is called AirshipConfig.plist and
    it is at the root level of the bundle

    :param application_path: Path to the .app application package
    :type key: str
    :param plist_name: Configuration plist file name, defaults to
        "AirshipConfig.plist"
    :returns: AirshipConfig.plist as a dictionary
    :rtype: dict
    """

    log.debug("Extracting {} at path {}".format(plist_name, application_path))
    if plist_name is DEFAULT_CONFIG_PLIST_NAME:
        airship_config_plist =\
            support.get_value_from_config(support.AIRSHIP_CONFIG_PLIST)
    else:
        airship_config_plist = plist_name

    path_to_plist = os.path.join(application_path, airship_config_plist)

    plist = None
    try:
        plist = plistlib.readPlist(path_to_plist)
        log.debug("AirshipConfig.plist:%s", plist)

    except IOError as error:
        if error.errno == errno.ENOENT:
            log.error(AIRSHIP_CONFIG_MISSING_MSG.format(application_path,
                                                        plist_name))
            return None

    except ExpatError:
        log.info(DEPRECATED_PLIST_CONVERT_MSG)
        plist = convert_deprecated_plist_at_path(path_to_plist)
        if plist is None:
            return None
        else:
            log.info("Plist conversion successful.")

    # Check for key change in LIB-608/DEVEXP-55, changes keys from CAMEL_CAPS
    # to lowerPascalCase for KVC property setting on iOS
    if plist.get("productionAppSecret") is None:
        plist = update_plist_keys_to_kvc_from_upper_snake(plist)

    return plist


def update_plist_keys_to_kvc_from_upper_snake(plist):
    """Take a Plist with deprecated keys and update it

    Changing from:
    APP_STORE_OR_AD_HOC = "APP_STORE_OR_AD_HOC_BUILD"
    DEVELOPMENT_APP_KEY = "DEVELOPMENT_APP_KEY"
    DEVELOPMENT_APP_SECRET = "DEVELOPMENT_APP_SECRET"
    PRODUCTION_APP_SECRET = "PRODUCTION_APP_SECRET"
    PRODUCTION_APP_KEY = "PRODUCTION_APP_KEY"

    To:
    APP_STORE_OR_AD_HOC_BUILD = "inProduction"
    DEVELOPMENT_APP_KEY = "developmentAppKey"
    DEVELOPMENT_APP_SECRET = "developmentAppSecret"
    PRODUCTION_APP_SECRET = "productionAppSecret"
    PRODUCTION_APP_KEY = "productionAppKey"

    :param plist: Plist with keys in the form CAPS_WITH_UNDERSCORE
    :return: Plist with KVC compliant keys, lowerCamelCase"""

    updated_plist = dict()
    updated_plist[APP_STORE_OR_AD_HOC_BUILD] = \
        plist[DEPRECATED_APP_STORE_OR_AD_HOC]
    updated_plist[DEVELOPMENT_APP_KEY] = \
        plist[DEPRECATED_DEVELOPMENT_APP_KEY]
    updated_plist[DEVELOPMENT_APP_SECRET] = \
        plist[DEPRECATED_DEVELOPMENT_APP_SECRET]
    updated_plist[PRODUCTION_APP_SECRET] = \
        plist[DEPRECATED_PRODUCTION_APP_SECRET]
    updated_plist[PRODUCTION_APP_KEY] = \
        plist[DEPRECATED_PRODUCTION_APP_KEY]
    return updated_plist


def convert_deprecated_plist_at_path(path_to_plist):
    """Extract a deprecated plist from an app adn convert it

    :param path_to_plist: Path to the plist to convert. System call PlistBuddy
    (/usr/libexec/PlistBuddy) is made, and that converts the plist

    :type key: str
    :return: Converted plist or None on error
    :rtype: dict
    """

    deprecated_plist_path = os.path.join(tempfile.mkdtemp(),
                                         CONVERTED_DEPRECATED_PLIST)
    plutil_command = ["/usr/bin/plutil", "-convert", "xml1", path_to_plist,
                      "-o", deprecated_plist_path]
    try:
        subprocess.check_call(plutil_command)
    except subprocess.CalledProcessError as error:
        log.error("plutil conversion failed with error code " + error.message)
        return

    try:
        plist = plistlib.readPlist(deprecated_plist_path)
    except ExpatError as error:
        log.error("""Cannot read converted plist, please check the format, and
                  possibly convert to the new style Error:%s""", error.message)
        return
    return plist

CODESIGN_PATH_ERROR = """
Check path to codesign tool with `which codesign`
and ensure your shell $PATH contains that path, you may
need to install the Command Line Tools
"""

CODESIGN_BAD_DIR_ERROR = """
Check the path to the directory and verify that it is an iOS Application
bundle.
"""

CODESIGN_NO_DIR_ERROR = """
Codesign verification failed, check the path to the app for errors
Path {path}
"""

CODESIGN_SUBPROC_ERROR = """
Error in codesign system call, please run with -d flag and email the diagnostic
output to support@urbanairship.com
"""

CODESIGN_TOOL_PATH_ERROR = """
The codesign tool doesn't appear to be on the shell path. Check it with
`which codesign`, and change the path. If you have further issues, please
run the tool with the -d flag and email the diagnostic output to
support@urbanairship.com
"""


def execute_codesign_system_call(application_path):
    """Executes /usr/bin/codesign, and captures part of the output.

    The other part of the output is written to the /tmp dir in the form of
    a plist. If there is a CalledProcessError, an attempt to check for the
    existence of the codesign tool by calling `which codesign`, followed by
    a check to see if the directory exists. If the directory exists, but the
    codesign tool fails, it is assumed that it is not an app bundle.

    :param application_path: Path to the .app application package
    :type key: str
    :returns: True if the codesign call was successful, False if not
    :rtype: bool
    """

    command = codesign_system_call_args(application_path)
    try:
        subprocess.check_call(command)
    except subprocess.CalledProcessError:
        # If the codesign tool is on the path, then directory must be off
        if not os.path.exists(application_path):
            log.error(CODESIGN_NO_DIR_ERROR.format(path=application_path))
            return False
        else:
            log.error(CODESIGN_BAD_DIR_ERROR)
            return False

    except OSError as os_error:
        if os_error.errno == errno.ENOENT:
            log.error(CODESIGN_TOOL_PATH_ERROR)
            return False
        else:
            log.error(CODESIGN_SUBPROC_ERROR)
            raise

    return True


def read_entitlement_plist_from_path(path_to_plist):
    """ Reads the plist at the given path.

    This will catch an IOError exception for a bad path and return None

    :param path_to_plist: Path to the entitlement plist generated by the
    codesign call. Default output is /tmp

    :type key: str
    :return: Contents of the plist
    :rtype: dict or None if there is an error
    """

    try:
        plist = plistlib.readPlist(path_to_plist)
        log.debug("Entitlement Plist:%s", plist)
    except IOError:
        log.error("""There is no plist at path %s, check the codesign command
                  setup""", path_to_plist)
        return None
    return plist


def codesign_system_call_args(path_to_app):
    """ Returns the codesign call with the -dvvv flag and args

    :param path_to_app:  Path to the .app file
    :type key: str
    :returns: List with proper command line options to execute as a system
    call

    :rtype: list
    """
    if path_to_app is None:
        raise TypeError("path_to_app cannot be None")

    plist_path = support.get_value_from_config(support.ENTITLEMENTS_PLIST_PATH)
    # Prepend the path with a colon for the system call
    plist_path = str.format(":{0}", plist_path)
    codesign_call = ["codesign", "-dvvv", "--entitlements", plist_path,
                     path_to_app]
    log.debug("Codesign system call:%s", codesign_call)
    return codesign_call

# Remedial error messages, makes console output cleaner

# check_apns_server error message
REMEDIAL_APS_ACTION = """
Your AirshipConfig.plist APP_STORE_OR_AD_HOC_BUILD key is set to {0}.
Your APS server in the entitlements file is {1}.
The UA server is set for of {2}.
All of these values need to match, you need to review these settings.

Change one or more of the following settings to reconfigure:

1. APP_STORE_OR_AD_HOC_BUILD key, currently -> {3}, {entitle_server}
2. Mobile provisioning file, currently configured for {4}
3. Change Urban Airship configuration, which is {ua_server}
4. Change the app key and secret, which are:

Key:{app_key}
Secret:{app_secret}
"""

# key_secret fail message
AIRSHIP_CONFIG_PARSE_FAIL = """
AirshipConfig.plist does not contain a bool or
string value for the APP_STORE_OR_AD_HOC key. Use PlistBuddy,
plutil, or Xcode to check the plist values"""

# check_bundle_id fail message
CHECK_BUNDLE_ID_FAIL = """Application bundle id {app_bundle} does not match
the bundle id configured on Urban Airship {ua_bundle}. Check the configuration
"""

# check_aps_environment fail message
APS_ENVIRONMENT_KEY_FAIL = """
No APS environment string, the provisioning profile was not created with
push enabled, rebuild the profile with proper entitlements by enabling push
for your app, or simply recreating the provisioning profile after checking the
app settings on the Apple Developer portal. For reference, the bundle id of
your app is:\n{bundle_id}"""


# Support classes
class VerificationInformation(object):
    """Model object for data associated with build and verification."""

    def __init__(self, entitlements_plist=None,
                 airship_config_plist=None):
        """ Initialize object with entitlements and AirshipConfig plists

        :param entitlements_plist: Plist returned from codesign -dvvv command
        :type key: dict
        :param airship_config_plist: AirshipConfig.plist has a dictionary
        :type key: dict
        """

        self.entitlements_plist = entitlements_plist
        self.airship_config_plist = airship_config_plist
        # Bundle id with no app id
        self.bundle_id = None

    def __str__(self):
        return "\nEntitlements:\n{0}\nAirshipConfig:\n{1}\n".format(
            self.entitlements_plist,
            self.airship_config_plist)

    def airship_config_is_production(self):
        """Parse the AirshipConfig.plist and return the  APP_STORE_OR_AD_HOC
        value.

        Because of legacy reasons, and to support bad configurations, the
        value stored in the AirshipConfig.plist for the APP_STORE_OR_AD_HOC
        key could be either a bool or a string. This handles both cases,
        and returns a bool. If the value is not identifiable as a bool or
        string, the tool exits with a non zero exit value.

        :returns: Value for the APP_STORE_OR_AD_HOC key
        :rtype: bool
        """

        is_production = self.airship_config_plist[APP_STORE_OR_AD_HOC_BUILD]
        # TODO
        # If is_production is not a bool, then this is a non xml plist, and
        # we look for a string YES/NO. Here we check for a string, and
        if not type(is_production) == bool:
            if type(is_production) == str:
                match = re.match('[Yy]', is_production)
                if match is None:
                    is_production = False
                else:
                    is_production = True
            else:
                # If not a bool or string, quit
                log.error(AIRSHIP_CONFIG_PARSE_FAIL)
                sys.exit(support.EXIT_FAILURE)
        return is_production

    def key_secret(self):
        """ Returns a named tuple with the key and secret

        Key and secret are parsed from the AirshipConfig.plist
        This will be either the production or development key/secret pair
        depending on the value of the APP_STORE_OR_AD_HOC_BUILD variable set
        in the plist.

        :returns: Named tuple with the values 'key','secret'
        :rtype: tuple
        """

        is_production = self.airship_config_is_production()

        if is_production:
            log.info("App is set for production in AirshipConfig")
            key = self.airship_config_plist[PRODUCTION_APP_KEY]
            secret = self.airship_config_plist[PRODUCTION_APP_SECRET]
        else:
            log.info("App is set for development in AirshipConfig")
            key = self.airship_config_plist[DEVELOPMENT_APP_KEY]
            secret = self.airship_config_plist[DEVELOPMENT_APP_SECRET]

        Auth = namedtuple('Auth', 'key secret')
        return Auth(key=key, secret=secret)

    def check_aps_environment(self):
        """APS environment check

        Check to see if the aps-environment keys exists

        :returns: True if the key exists, false otherwise
        :rtype: bool
        """
        try:
            full_bundle_id = self.entitlements_plist[ENTITLE_APP_ID_KEY]
        except KeyError:
            log.error("Key error")
            return False

        try:
            self.entitlements_plist[ENTITLE_APS_ENV_KEY]
        except KeyError:
            remedial = APS_ENVIRONMENT_KEY_FAIL.format(bundle_id=full_bundle_id)
            log.error(remedial)
            return False

        msg = "Environment is set properly for bundle id {0}".format(
            full_bundle_id)
        log.info(msg)
        return True

    def check_apns_server(self, ua_verification_response):
        """ Check the production/development server in entitlements

        The API call to get data was authenticated with the key/secret pair
        related to the APP_STORE_OR_AD_HOC_BUILD value in the AirshipConfig
        plist. The API returns production or development based on those keys,
        this test checks that value against the value in the entitlements
        for the app.

        :param ua_verification_response: VerificationInformation object
        associated with this operation.

        :type key: str
        :returns: True if the servers match, false otherwise
        :rtype: bool
        """

        adhoc_or_prod = self.airship_config_is_production()

        if adhoc_or_prod:
            aps_env_description = PRODUCTION_SERVER
            app_key = self.airship_config_plist[PRODUCTION_APP_KEY]
            app_secret = self.airship_config_plist[PRODUCTION_APP_SECRET]
        else:
            aps_env_description = DEVELOPMENT_SERVER
            app_key = self.airship_config_plist[DEVELOPMENT_APP_KEY]
            app_secret = self.airship_config_plist[DEVELOPMENT_APP_SECRET]

        if ua_verification_response[UA_API_RESPONSE_PRODUCTION_KEY]:
            ua_aps_configuration = PRODUCTION_SERVER
        else:
            ua_aps_configuration = DEVELOPMENT_SERVER

        aps_environment = self.entitlements_plist[ENTITLE_APS_ENV_KEY]

        # Line up the servers
        if not ua_aps_configuration == aps_environment == aps_env_description:
            # TODO fix these key/values up, so they are no more numbers or
            # repetition, I mean, this is just bad.
            remedial = REMEDIAL_APS_ACTION.format(aps_env_description,
                                                  aps_environment,
                                                  ua_aps_configuration,
                                                  adhoc_or_prod,
                                                  aps_environment,
                                                  ua_aps_configuration,
                                                  app_key=app_key,
                                                  app_secret=app_secret,
                                                  ua_server=ua_aps_configuration,
                                                  entitle_server=aps_env_description)
            log.error(remedial)
            return False

        msg = "\nUrban Airship server is configured for: {0}\n".format(
            ua_aps_configuration)
        msg += "App is configured for: {0}".format(aps_environment)
        log.info(msg)
        return True

    def check_bundle_id(self, ua_verification_response):
        """Check the bundle id for errors

        The API call returns the bundle ID associated with the app
        authenticated with the key/secret. Keep in mind, the bundle id in the
        entitlements is prepended with the App ID

        :param ua_verification_response: JSON response from UA API for this
        operation

        :type key: dict
        :returns: True is the bundle id is correct, false otherwise
        :rtype: bool
        """

        entitlements_bundle_id = self.entitlements_plist[ENTITLE_APP_ID_KEY]
        entitlements_bundle_id = entitlements_bundle_id.split('.')
        entitlements_bundle_id = entitlements_bundle_id[1:]
        entitlements_bundle_id = string.join(entitlements_bundle_id, '.')
        ua_bundle_id = ua_verification_response[UA_API_RESPONSE_BUNDLE_ID_KEY]
        log.info("Application Bundle id is: %s", entitlements_bundle_id)
        log.info("Urban Airship Configured App Bundle id is: %s", ua_bundle_id)

        if entitlements_bundle_id == ua_bundle_id:
            log.info("Bundle id is correct")
            return True
        else:
            remedial = CHECK_BUNDLE_ID_FAIL.format(
                app_bundle=entitlements_bundle_id, ua_bundle=ua_bundle_id)
            log.error(remedial)
            return False

    def log_diagnostic_information(self):
        """Log out entitlements/config information for diagnostics
        """

        log.debug("ENTITLEMENTS:\n%s", self.entitlements_plist)
        log.debug("AIRSHIP_CONFIG:\n%s", self.airship_config_plist)
        log.debug("UA_API_RESPONSE:\n%s", support._ua_api_response)



"""Common functionality for iOS and Android verification"""

#    Copyright 2012 Urban Airship
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import ConfigParser
import json
import logging
import urllib2
import urlparse
import base64
import sys
import os

from datetime import datetime
from collections import namedtuple

# Logging name
UA_LOGGER = "com.urbanairship.uaverify"

# UA API URLS
UA_API_BASE = "api_base"
UA_API_PATH = "api_path"

# Exit status
EXIT_SUCCESS = 0
EXIT_FAILURE = 1

# Configuration values
DEFAULT_CONFIG_NAME = "ua_inspection.cfg"
UA_SECTION = "ua_config"
LOG_LEVEL = "log_level"
ENTITLEMENTS_PLIST_PATH = "entitlement_plist_path"
AIRSHIP_CONFIG_PLIST = "airship_config_plist"

# Response for UA API requests
Response = namedtuple('Response', ["json", "error"])

# Configuration parsing
# Load configuration file, or log an error
config_parser = ConfigParser.ConfigParser()
config_path = os.path.dirname(os.path.abspath(__file__))
config_path = os.path.join(config_path, DEFAULT_CONFIG_NAME)

if not os.path.exists(config_path):
    logging.error("No config file found at path %s", config_path)
    sys.exit(EXIT_FAILURE)

config_parser.read(config_path)

# Store the API response to use in diagnostic print outs
_ua_api_response = None


def get_base_verification_request(appKey, appSecret):
    """Get the verification request.

    The request is setup with
    basic auth headers authenticated with the appKey and appSecret

    :param appKey: The application key
    :type key: str
    :param appSecret: The application secret
    :type key: str
    :return: Pre authenticated http request
    :rtype: urllib2.Request
    """

    base_url = get_value_from_config(UA_API_BASE)
    api_url = get_value_from_config(UA_API_PATH)
    ua_url = urlparse.urljoin(base_url, api_url)
    log.info("URL for verification call %s", ua_url)
    basic_auth = "{0}:{1}".format(appKey, appSecret)
    basic_auth_encoded = base64.encodestring(basic_auth)
    pre_auth = {'Authorization': "Basic %s" % basic_auth_encoded}
    log.debug("Headers for UA verification request %s", pre_auth)
    return urllib2.Request(ua_url, headers=pre_auth)


def get_value_from_config(key, section=UA_SECTION):
    """Returns a configuration value for the given key.

    Return the value associated with the key
    from the configuration file. Uses the default
    section defined by UA_SECTION unless overridden
    See ConfigParser for more details

    :param key: Key associated with value
    :type key: str
    :param section: Section to retrieve value from
    :type section: str
    """

    return config_parser.get(section, key)


# Logging setup

# Timestamp and log formatting
# LogLevel
log_format = '%(asctime)s:%(levelname)s %(message)s'
logging.basicConfig(format=log_format)
log = logging.getLogger(UA_LOGGER)
log.setLevel(get_value_from_config(LOG_LEVEL))


def make_request_against_api(request):
    """Make a request against the UA API

    Returns a named tuple with response data as JSON and a possible error
    object which is either an HTTPError, 401,500, etc or a URLError which
    most likely occurs when there is no connection. All other errors will
    exception out.

    :param request: Request to make
    :type key: urllib2.Request
    :return: Named tuple in the form (response, error)
    :rtype: namedtuple
    """

    error = None
    json_data = None

    try:
        raw_response = urllib2.urlopen(request)
        json_data = json.load(raw_response)

    except urllib2.HTTPError as http_error:
        log.error("HTTP error %s with code %d", http_error.message,
                  http_error.code)
        error = http_error

    except urllib2.URLError as url_error:
        log.error("URL error, are you connected to the internet?")
        log.error("URL error message %s", url_error)
        error = url_error

    # Saving UA API response in case diagnostic logging is used.
    _ua_api_response = Response(json=json_data, error=error)
    return _ua_api_response


def setup_diagnostic_logging():
    """Adds a file handler to the logger object to output information

    Also attempts to prepend the diagnostic log file with entitlement and
    configuration information
    """

    log.setLevel(logging.DEBUG)
    now = datetime.now()
    filename = now.strftime("ua_verify_diagnostic_%d%b%y_%X.txt")
    path = os.getcwd()
    file_path = os.path.join(path, filename)
    log.info("Diagnostic log path %s", file_path)
    log_handler = logging.FileHandler(file_path)
    log_handler.setLevel(logging.DEBUG)
    log.addHandler(log_handler)
